#!/bin/bash

##############################################################################
#
# Juice4halt setup script
#
# TAG Juice4HaltSetup.sh
# Version 1.2.1 - 2019-05-28
# - cor: passing an empty argument now correctly defaults to "auto".
#
# TAG Juice4HaltSetup.sh
# Version 1.2.0 - 2019-05-22
# - mod: now uses a single script for start-stop.
#
# TAG Juice4HaltSetup.sh
# Version 1.1.0 - 2019-05-22
# - add: script will now terminate early if no signal received from board for a number of seconds.
# - mod: remodeled some functions and improved the help.
#
# TAG Juice4HaltSetup.sh
# Version 1.0.0 - 2019-05-21
# - first release
#
################################################################################


# for help, type: sudo ./Juice4HaltSetup help

# Board status
# GPIO25 toggles the juice4halt board on/off
#
# When the board is inactive, it will not send any signal and will not perform an automatic shutdown.
# It will simply let power drain and the RPi will eventually stop.
#
# When active, it will send a 700ms low signal when power has been lost for 20 seconds.
# If not signal is received back, It will simply let power drain and the RPi will eventually stop.
# If it receives a low signal, it will continue powering the RPi to allow for a clean shutdown.
# When the GPIO25 state is back to high, which is done automatically when the RPi is off as the pin is released,
# the board resets to its inital state.
#
# Thus, the board should be activated at boot and deactivated at reboot (or if the service is stopped).
# At shutdown, the pin is pulled low and remains low until the RPi is shutdown.


# juice4halt.service
# The heart of the scripts is juice4halt.service, a systemd service containing 2 scripts:

# j4hmonitor:
# - started at boot (or when service is started manually)
# - constantly running
# - activates the board
# - monitor for power outage
# - in case of power failure, deactivates the board, perform shutdown and stops


# GPIO 25
# -------
# Use gpio 25 as an open-drain output.
# Send a short LOW pulse = Booting finished, monitor script runs, automatic shutdown enabled,
# then change to input and listen.
# If a LOW is sent from the board then initiate shutdown.

# GPIO22
# -------
# Use gpio 22 as an open drain output to detect power failure immediately.
# To read the value, set it to output, set it to high and read the value back.
# Its value will immediately be pulled back low if power has failed.




##############################################################################
# SCRIPT AND TEXT BLOCK FUNCTIONS
##############################################################################


# -- getScriptMonitor

function getScriptMonitor {

cat <<"EndOfScript"
#!/bin/bash

#
# title: j4hmonitor
# description: This script runs on Raspberry Pi equipped with the Juice4Halt module.
# original author: Nelectra s.r.o.
# adapted for systemd by: Bruno Frechette (Newmood)
# notes: www.juice4halt.com
# copyright: GNU GPL v3.0
#


# -- log functions

function logEvent {

	# define log file
	local logFile="/usr/local/bin/juice4halt/j4h.log"

	# define the script name to add to log lines
	local logScriptName="$(basename "${0}")"

	# always log to system log
	logger --tag "juice4halt[${logScriptName}]" "${1}"

	if [[ $isLogEnabled != "true" ]] ; then return ; fi

	# format date
    local datestring=`date +'%Y-%m-%d %H:%M:%S'`

    # Write to local log file
	echo "${datestring} ${logScriptName}: ${1}" >> "${logFile}"

	# maintain the length of the log file to a reasonable length
	# when it reaches 1500 lines, truncate it to 1000
	if [[ $(wc -l < "${logFile}") -gt 1500 ]] ; then
		echo "$(tail -1000 ${logFile})" > "${logFile}"
	fi
	}


function logPowerStatus {
	if [[ "${1}" == "0" ]] ; then
		logEvent "power failure - RPi powered by juice4halt board"
	else
		logEvent "power restored - RPi powered by main power input"
	fi
	}


function getElapsedSeconds {
	cat /proc/uptime | cut --delimiter=' ' --field=1 | cut --delimiter=. --field=1
	}

function gpioRead { cat "/sys/class/gpio/gpio${1}/value" ; }
function gpioReadDirection { cat "/sys/class/gpio/gpio${1}/direction" ; }
function gpioWriteLow { echo "0" > "/sys/class/gpio/gpio${1}/value" ; }
function gpioWriteHi { echo "1" > "/sys/class/gpio/gpio${1}/value" ; }
function gpioSetIn { echo "in" > "/sys/class/gpio/gpio${1}/direction" ; }
function gpioSetOut { echo "out" > "/sys/class/gpio/gpio${1}/direction" ; }
function gpioExport { echo "${1}" > "/sys/class/gpio/export" ; sleep 0.2s ; }
function gpioInit { if [[ ! -f "/sys/class/gpio/gpio${1}/value" ]] ; then gpioExport "${1}" ; fi ; }

function startService {

	# setup GPIO25 used to communicate board status
	logEvent "activating board (GPIO25 100ms low pulse)"
	gpioInit "25"
	gpioSetOut "25"
	gpioWriteLow "25"
	sleep 0.1s

	# setup GPIO22 used to monitor input power
	gpioInit "22"
	gpioSetOut "22"

	# set back to input, wait until pin rises to HI
	gpioSetIn "25"
	sleep 0.1s

	# initialize gpio read variables
	g25val1="1"
	g25val2="1"
	g22val1="1"
	g22val2="1"
	
	# initialize variables
	previousBoardStatus="1"
	currentBoardStatus="1"
	boardStatusLowStart=$(getElapsedSeconds)
	previousPowerStatus="1"
	currentPowerStatus="1"


	# to be resistant to short pulses
	# (e.g. 100ms pulse generated by the stop action)
	# two readings 200ms apart from each other are required

	# also, to avoid rebooting if the card is no longer present,
	# the power input must also indicate that main power is out
	# the script will auto-exit if pin 25 is held low for over 10 seconds

	while [ "${currentBoardStatus}" == "1" -o "${currentPowerStatus}" == "1" ]
	do
			# perform the check every 200 ms to reduce cpu usage
			# any shutdown request will leave the pin low for a 700 ms
			sleep 0.2s


			# -- read gpio

			# first read
			g25val1=$(gpioRead "25")
			gpioWriteHi "22"
			g22val1=$(gpioRead "22")

			# wait
			sleep 0.2s

			# second read
			g25val2=$(gpioRead "25")
			gpioWriteHi "22"
			g22val2=$(gpioRead "22")


			# -- update data

			# only update gpio data if both readings are the same
			# otherwise keep current values
			if [[ "${g25val1}" == "${g25val2}" ]] ; then currentBoardStatus="${g25val1}" ; fi
			if [[ "${g22val1}" == "${g22val2}" ]] ; then currentPowerStatus="${g22val1}" ; fi


			# -- process board status

			# if gpio25 goes low, record time of change
			if [[ "${currentBoardStatus}" == "0" && "${previousBoardStatus}" == "1" ]] ; then boardStatusLowStart=$(getElapsedSeconds) ; fi

			# check if gpio25 has been low for over 120 seconds
			# if so, loop was never exited because gpio22 was never pulled low
			# this probably means that there is no board present
			if [[ "${currentBoardStatus}" == "0" ]] ; then

				elapsedTime=$(($(getElapsedSeconds) - ${boardStatusLowStart}))

				if [ ${elapsedTime} -gt 120 ] ; then
					logEvent "*** no power signal received from board for over ${elapsedTime} seconds"
					logEvent "*** board presumed absent - exiting script"
					exit
				fi
			fi


			# -- process power status

			# log change in power status
			if [[ "${currentPowerStatus}" != "${previousPowerStatus}" ]] ; then logPowerStatus "${currentPowerStatus}" ; fi


			# -- update internal status

			previousBoardStatus="${currentBoardStatus}"
			previousPowerStatus="${currentPowerStatus}"

	done

	logEvent "*** shutdown requested by board because of power failure"

	# set GPIO25 as output and set output to LOW
	# this signals the board that the shutdown is initiated
	# the board will not cut power to the RPi until the output is back to HIGH
	# after system halted GPIO25 will be automatically switched to input and the level will be pulled up to HI
	# a LOW to HI transition signals to the board that clean shutdown is completed
	# the board will then cut power to the RPi and be ready to restart when power is restored

	logEvent "signaling board that shutdown is in progress (GPIO25 set low)"
	gpioSetOut "25"
	gpioWriteLow "25"

	logEvent "shutting down RPi"
	shutdown -h now
	}


function stopService {

	# Before signaling the board, check if gpio25 is already held low.
	# If so, a shutdown is in progress and sending a pulse on gpio25
	# will be interpreted as "shutdown finished" and the board will
	# immediately cut power to the RPi.

	# read current values
	gpio25direction=$(gpioReadDirection "25")
	gpio25value=$(gpioRead "25")

	# exit if shutdown in progress
	if [[ "${gpio25direction}" == "out" && "${gpio25value}" == "0" ]] ; then
		logEvent "shutdown in progress - no signal sent to board"
		exit
	fi

	# Use the pin as an open-drain output.
	# Send a short LOW pulse to the board to indicate "deactivate automatic shutdown".
	# It will be reactivated during the boot process.

	logEvent "deactivating board (GPIO25 100ms low pulse)"

	# set GPIO25 as output and set output to LOW
	gpioSetOut "25"
	gpioWriteLow "25"
	sleep 0.1s

	# change back to input, wait until pin rises to HI
	gpioSetIn "25"
	sleep 0.1s

	}

# -- main

# get action: start | stop
ScriptAction="${1}"

# set to false to disable logging in local file
# logging is always enabled in system log facility
isLogEnabled=true

logEvent "received signal ${ScriptAction}"

case "${ScriptAction}" in

	"start")
		startService
		;;

	"stop")
		stopService
		;;

	*)
		logEvent "Unrecognized command: ${ScriptAction}"
		;;

esac


exit 0


EndOfScript

}


# -------------------------------------------------------------------------------------


# -- getServiceFile

function getServiceFile {

cat <<"EndOfScript"
[Unit]
Description=Juice4Halt service

[Service]
Type=simple
ExecStart=/usr/local/bin/juice4halt/j4hmonitor "start"
ExecStop=/usr/local/bin/juice4halt/j4hmonitor "stop"

[Install]
WantedBy=default.target

EndOfScript

}


# -------------------------------------------------------------------------------------


# -- getTextHelp

function getTextHelp {

cat <<"EndOfMessage"
NAME
	Juice4HaltSetup - installation script for the Juice4Halt board

SYNOPSIS
	Juice4HaltSetup [COMMAND]

DESCRIPTION
	Install or remove the required scripts for the Juice4Halt card.
	The script must be run as root.
	
COMMANDS
	
	auto, no command
		automatically detect the card and perform either installs or uninstalls the scripts
		passing an empty command is the equivalent of the "auto" command

	detect
		detects if the juice4halt card is physically installed
		
	install
		install the juice4halt service and script even if the card is not detected

	uninstall
		uninstall the juice4halt scripts and service

	help
		print this help

EXAMPLES
	sudo ./Juice4HaltSetup
	sudo ./Juice4HaltSetup uninstall

FILES
	Setup script: Juice4HaltSetup
	Main script: /usr/local/bin/juice4halt/j4hmonitor
	Log file: /usr/local/bin/juice4halt/jh4.log
	Systemd service file: /lib/systemd/system/juice4halt.service

AUTHOR
	Original author: Nelectra s.r.o.
	Modified and adapted for systemd by Bruno Frechette (Newmood)

COPYRIGHT
	License GNU GPL v3.0
	This is free software: you are free to change and redistribute it.
	There is NO WARRANTY, to the extent permitted by law.

SEE ALSO
	Full documentation on the Juice4halt board at: <http://juice4halt.com>

EndOfMessage

}





##############################################################################
# TEXT FORMAT FUNCTIONS
##############################################################################



# ----- echoHashline ----- #

function echoHashline {
	echo "################################################################################"
}



# ----- echoHashBorder ----- #

function echoHashBorder {
	echo "#                                                                              #"
}



# ----- echoHashCenter ----- #

function echoHashCenter {
	local tText="${1}"
	local tNumChars=${#tText}
	local tLeftPadLength=$((39-$tNumChars/2))
	local tRightPadLength=$((78-$tNumChars-$tLeftPadLength))
	local tSpaces="                                        "
	echo "#${tSpaces:0:tLeftPadLength}${tText}${tSpaces:0:tRightPadLength}#"
}



# ----- echoCenter ----- #

function echoCenter {
	local tText="${1}"
	local tNumChars=${#tText}
	local tPadLength=$((40-$tNumChars/2))
	local tSpaces="                                        "
	echo "${tSpaces:0:$tPadLength}${tText}"
}



# ----- echoHashBox ----- #

function echoHashBox {
	local tText="${1}"
	echoHashline
	echoHashBorder
	echoHashCenter "${tText}"
	echoHashBorder
	echoHashline
	echo
}






##############################################################################
# INSTALLER DISPLAY FUNCTIONS
##############################################################################

# ----- showErrorExit ----- #

function showErrorExit {
	echo "*** ERROR *** ${1}"
	echo ""
	exit 1
}



# ----- showEvent ----- #

function showEvent {
	echo "- ${1}"
}



# ----- showMajorEvent ----- #

function showMajorEvent {
	echo ""
	echo "*** ${1} ***"
	echo ""
}






##############################################################################
# UTILITY FUNCTIONS
##############################################################################


# ----- checkForRoot ----- #

function checkForRoot {
	if [[ $EUID > 0 ]] ; then
		showErrorExit "Please run as root"
	fi
}



# ----- checkForSystemd ----- #

# this is a very basic function to detect the presence of the /etc/systemd directory
# it is by no means a comprehensive test

function checkForSystemd {
	if [ ! -d "/etc/systemd" ] ; then
		showErrorExit "This script is only compatible with distributions supporting systemd"
	fi
}



# ----- detectCardGPIO ----- #

function detectCardGPIO {

	# detect the card using the gpio number passed in argument
	# this is a crude method and could give a false positive
	# if another card is installed and uses the same pin
	# the argument is the GPIO number, not the physical pin
	# if the card is connected and powered on, the value of the pin should be high
	# returns 0 if the card is not detected (pin low)
	# returns 1 if the card is detected (pin high)

	local gpioNo="${1}"

	# activate the gpio pin
	if [[ ! -f "/sys/class/gpio/gpio${gpioNo}/value" ]] ; then
		echo "${gpioNo}" > /sys/class/gpio/export
		sleep 0.2s
	fi

	# set the pin as input
	echo "in" > "/sys/class/gpio/gpio${gpioNo}/direction"
	sleep 0.2s

	local pinValue
	pinValue=$(cat "/sys/class/gpio/gpio${gpioNo}/value")

	if [[ $pinValue -eq 0 ]] ; then
		showEvent "no card detected on gpio ${gpioNo}..."
		return 0
	else
		showEvent "card detected on gpio ${gpioNo}!"
		return 1
	fi
}



# ----- signalGPIO ----- #

function signalGPIO {

	# arguments: gpioNo pinValue

	local gpioNo="${1}"
	local pinValue="${2}"

	# only send signal if the pin has already been exported
	if [[ ! -f "/sys/class/gpio/gpio${gpioNo}/value" ]] ; then
		return
	fi

	# set GPIO pin as output and set output to passed value
	echo "out" > "/sys/class/gpio/gpio${gpioNo}/direction"
	echo "${pinValue}" > "/sys/class/gpio/gpio${gpioNo}/value"
	sleep 0.1s

	# set GPIO pin to input and wait
	echo "in" > "/sys/class/gpio/gpio${gpioNo}/direction"
	sleep 0.1s
}




##############################################################################
# FILE FUNCTIONS
##############################################################################

# ----- createDir ----- #

function createDir {
	# usage: createdir /path/to/directory

	local tDir="${1}"

	if [ -d "${tDir}" ] ; then
		showEvent "directory ${tDir} already exists"
		return 0
	fi

	mkdir "${tDir}"

	if [[ $? -ne 0 ]] ; then showErrorExit "could not create the directory ${tDir}" ; fi
	if [ ! -d "${tDir}" ] ; then showErrorExit "could not create the directory ${tDir}" ; fi

	showEvent "directory ${tDir} created"

}



# ----- createFile ----- #

function createFile {
	# usage: createFile /path/to/file filecontent

	local tFilepath="${1}"
	local tContent="${2}"
	local tAction="created"

	if [ -f "${tFilepath}" ] ; then	tAction="replaced" ; fi

	echo "${tContent}" > ${tFilepath}

	if [[ $? -ne 0 ]] ; then showErrorExit "could not create the reboot script file ${tFilepath}" ; fi
	if [ -f "${tFilepath}" ] ; then
		showEvent "file ${tFilepath} ${tAction}"
	else
		showErrorExit "could not create the file ${tFilepath}"
	fi
}



# ----- createFileExecutable ----- #

function createFileExecutable {
	# usage: createFile /path/to/file filecontent
	createFile "${1}" "${2}" 
	chmod +x "${1}"
}



# ----- removeFile ----- #

function removeFile {

	local tFilepath="${1}"
	if [[ ! -f "${tFilepath}" ]] ; then return 0 ; fi
	rm "${tFilepath}"

	if [[ $? -ne 0 ]] ; then
		showErrorExit "could not remove file ${tFilepath}"
	else
		showEvent "file removed : ${tFilepath}"
	fi
}



# ----- removeDirectory ----- #

function removeDirectory {

	local tLinkDir="${1}"
	if [ ! -d "${tLinkDir}" ] ; then return 0 ; fi

	rmdir "${tLinkDir}"

	if [[ $? -ne 0 ]] ; then
		showErrorExit "could not remove directory ${tLinkDir}"
	else
		showEvent "directory removed : ${tLinkDir}"
	fi
}


##############################################################################
# SYSTEMCTL FUNCTIONS
##############################################################################


# ----- isServiceActive ----- #

function isServiceActive {

	# systemctl is-active exit code is 0 if the system is active

	local tServiceName="${1}"

	systemctl --quiet is-active "${tServiceName}"
	local res=$?

	if [[ $res -eq 0 ]] ; then
		showEvent "${tServiceName} is currently active (running)"
		return 0
	else
		showEvent "${tServiceName} is currently not active (not running)"
		return 1
	fi

}



# ----- isServiceEnabled ----- #

function isServiceEnabled {

	# systemctl is-enabled exit code is 0 if the system is enabled

	# is-enabled prints an error message if the service is not found
	# since we only care about the exit code
	# to avoid the message being printed, redirect standard error to null

	local tServiceName="${1}"

	systemctl --quiet is-enabled "${tServiceName}" 2> /dev/null
	local res=$?

	if [[ $res -eq 0 ]] ; then
		showEvent "${tServiceName} is currently enabled (will start at boot)"
		return 0
	else
		showEvent "${tServiceName} is currently not enabled (will not start at boot)"
		return 1
	fi

	if [[ $res -eq 0 ]] ; then return 1 ; fi

}



# ----- disableService ----- #

function disableService {

	# usage: disableService serviceName

	local tServiceName="${1}"

	# unload service if already loaded
	if isServiceEnabled  "${tServiceName}" ; then

		showEvent "disabling ${tServiceName}..."
		systemctl --quiet disable "${tServiceName}"
		if [[ $? -ne 0 ]] ; then showErrorExit "could not disable service ${tServiceName}" ; fi
		showEvent "${tServiceName} disabled (will no longer start at boot)"

		# refresh systemd daemon
		systemctl daemon-reload

	fi

}



# ----- enableService ----- #

function enableService {

	# usage: enableService serviceName

	local tServiceName="${1}"

	# load service if not already loaded
	if ! isServiceEnabled  "${tServiceName}" ; then

		showEvent "enabling ${tServiceName}..."
		systemctl --quiet enable "${tServiceName}"
		if [[ $? -ne 0 ]] ; then showErrorExit "could not enable service ${tServiceName}" ; fi
		showEvent "${tServiceName} enabled (will start at boot)"

		# refresh systemd daemon
		systemctl daemon-reload

	fi


}



# ----- stopService ----- #

function stopService {

	# usage: stopService serviceName

	local tServiceName="${1}"

	# halt service if running
	if isServiceActive "${tServiceName}" ; then

		showEvent "stopping ${tServiceName}..."
		systemctl --quiet stop "${tServiceName}"
		if [[ $? -ne 0 ]] ; then showErrorExit "could not stop service ${tServiceName}" ; fi
		showEvent "${tServiceName} stopped"

	fi

}

# ----- startService ----- #

function startService {

	# usage: startService serviceName

	local tServiceName="${1}"

	# start service if not running
	if ! isServiceActive "${tServiceName}" ; then

		showEvent "starting ${tServiceName}..."
		systemctl --quiet start "${tServiceName}"
		if [[ $? -ne 0 ]] ; then showErrorExit "could not start service ${tServiceName}" ; fi
		showEvent "${tServiceName} now running"

	fi

}

##############################################################################
# MAIN PROCESS FUNCTIONS
##############################################################################


# ----- processInstall ----- #

function processInstall {


	showMajorEvent "checking if service is currently running"

	# stop and disable service in case they are running
	stopService "juice4halt.service"
	disableService "juice4halt.service"

	showMajorEvent "creating scripts"

	# create directory that will hold the scripts
	createDir "${ScriptDir}"

	# remove old log file in case of reinstallation
	removeFile "${ScriptDir}/j4h.log"

	# create monitor script file
	createFileExecutable "${ScriptDir}/j4hmonitor" "$(getScriptMonitor)"

	showMajorEvent "creating service file"

	# service files
	createFile "${ServiceFile}" "$(getServiceFile)"

	showMajorEvent "starting service"

	# start and enable service
	enableService "juice4halt.service"
	startService "juice4halt.service"

	showMajorEvent "Installation completed. Juice4Halt is now running. (reboot not required)"

}



# ----- processUninstall ----- #

function processUninstall {


	showMajorEvent "stopping service"

	# stop and disable service
	stopService "juice4halt.service"
	disableService "juice4halt.service"
	removeFile "${ServiceFile}"

	showMajorEvent "removing script files"

	removeFile "${ScriptDir}/j4hmonitor"
	removeFile "${ScriptDir}/j4h.log"
	removeDirectory "${ScriptDir}"

	showMajorEvent "Uninstallation completed. (reboot not required)"

}


# ----- processAuto ----- #

function processAuto {

	showMajorEvent "detecting card"
	detectCardGPIO "25"

	if [[ $? -ne 0 ]] ; then
		showEvent "card detected: installing service"
		processInstall
	else
		showEvent "no card detected: uninstalling service"
		processUninstall
	fi

}




##############################################################################
# MAIN
##############################################################################


# check requirements

checkForRoot
checkForSystemd


# set global variables

ScriptDir="/usr/local/bin/juice4halt"
ServiceFile="/lib/systemd/system/juice4halt.service"


# perform action according to script argument

case "${1}" in

	"" | "auto")
		echoHashBox "Juice4Halt setup script"
		processAuto
		;;


	"detect")
		echoHashBox "Juice4Halt setup script"
		showMajorEvent "Detecting card..."
		detectCardGPIO "25"
		;;


	"install")
		echoHashBox "Juice4Halt setup script"
		processInstall
		;;


	"uninstall")
		echoHashBox "Juice4Halt setup script"
		processUninstall
		;;


	"help")
		echo "$(getTextHelp)"
		;;


	*)
		showMajorEvent "Unrecognized argument: ${1}"
		echo "Type ${0} help for help"
		echo ""
		exit 1
		;;

esac


exit 0
